//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      Base/Axis/FixedBinAxis.cpp
//! @brief     Implement class FixedBinAxis.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "Base/Axis/FixedBinAxis.h"
#include "Base/Axis/Bin.h"
#include "Base/Util/Algorithms.h"
#include <iomanip>
#include <limits>

FixedBinAxis::FixedBinAxis(const std::string& name, size_t nbins, double start, double end)
    : IAxis(name)
    , m_nbins(nbins)
    , m_start(start)
    , m_end(end)
{
}

FixedBinAxis* FixedBinAxis::clone() const
{
    auto* result = new FixedBinAxis(axisName(), m_nbins, m_start, m_end);
    return result;
}

double FixedBinAxis::operator[](size_t index) const
{
    if (index >= m_nbins)
        throw std::runtime_error("FixedBinAxis::operator[] -> Error. Wrong index.");

    double step = (m_end - m_start) / m_nbins;
    return m_start + (index + 0.5) * step;
}

Bin1D FixedBinAxis::bin(size_t index) const
{
    if (index >= m_nbins)
        throw std::runtime_error("FixedBinAxis::bin -> Error. Wrong index.");

    double step = (m_end - m_start) / m_nbins;
    Bin1D result(m_start + step * index, m_start + step * (index + 1));
    return result;
}

size_t FixedBinAxis::findClosestIndex(double value) const
{
    if (value < min())
        return 0;
    if (value >= max())
        return m_nbins - 1;

    double step = (m_end - m_start) / m_nbins;
    return int((value - m_start) / step);
}

std::vector<double> FixedBinAxis::binCenters() const
{
    std::vector<double> result;
    result.resize(size(), 0.0);
    for (size_t i = 0; i < size(); ++i)
        result[i] = bin(i).center();
    return result;
}

std::vector<double> FixedBinAxis::binBoundaries() const
{
    std::vector<double> result;
    result.resize(size() + 1, 0.0);
    for (size_t i = 0; i < size(); ++i)
        result[i] = bin(i).m_lower;
    result[size()] = bin(size() - 1).m_upper;
    return result;
}

void FixedBinAxis::clip(double lower, double upper)
{
    if (lower >= upper)
        throw std::runtime_error(
            "FixedBinAxis::clip -> Error. 'lower' should be smaller than 'upper'");

    if (lower < min())
        lower = bin(0).center();
    if (upper >= max())
        upper = bin(size() - 1).center();

    const size_t nbin1 = findClosestIndex(lower);
    const size_t nbin2 = findClosestIndex(upper);

    // create tmp vars until everything is calculated, otherwise the calculation will be corrupted
    // by partially changed values
    const auto newStart = bin(nbin1).m_lower;
    const auto newEnd = bin(nbin2).m_upper;

    m_nbins = nbin2 - nbin1 + 1;
    m_start = newStart;
    m_end = newEnd;
}

void FixedBinAxis::print(std::ostream& ostr) const
{
    ostr << "FixedBinAxis(\"" << axisName() << "\", " << size() << ", "
         << std::setprecision(std::numeric_limits<double>::digits10 + 2) << min() << ", " << max()
         << ")";
}

bool FixedBinAxis::equals(const IAxis& other) const
{
    if (!IAxis::equals(other))
        return false;
    if (const auto* otherAxis = dynamic_cast<const FixedBinAxis*>(&other)) {
        if (size() != otherAxis->size())
            return false;
        if (!BaseUtils::algo::almostEqual(m_start, otherAxis->m_start))
            return false;
        if (!BaseUtils::algo::almostEqual(m_end, otherAxis->m_end))
            return false;
        return true;
    }
    return false;
}

//... In global namespace.

FixedBinAxis FixedScanAxis(const std::string& name, size_t nbins, double scan_start,
                           double scan_end)
{
    if (nbins < 2)
        throw std::runtime_error("Cannot create scan axis with less than two points");
    const double delta = (scan_end - scan_start) / (nbins - 1);
    return FixedBinAxis(name, nbins, scan_start - delta / 2, scan_end + delta / 2);
}
