//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      Base/Util/StringUtils.cpp
//! @brief     Implements a few helper functions
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "Base/Util/StringUtils.h"
#include <boost/algorithm/string.hpp>
#include <charconv>
#include <regex>

//! Returns string right-padded with blanks.
std::string BaseUtils::String::padRight(const std::string& name, size_t length)
{
    std::string result = name;
    result.resize(length, ' ');
    return result;
}

//! Returns token vector obtained by splitting string at delimiters.
std::vector<std::string> BaseUtils::String::split(const std::string& text,
                                                  const std::string& delimiter)
{
    std::vector<std::string> tokens;
    boost::split(tokens, text, boost::is_any_of(delimiter));
    return tokens;
}

void BaseUtils::String::replaceItemsFromString(std::string& text,
                                               const std::vector<std::string>& items,
                                               const std::string& replacement)
{
    for (size_t i = 0; i < items.size(); ++i)
        boost::replace_all(text, items[i], replacement);
}

std::string BaseUtils::String::join(const std::vector<std::string>& joinable,
                                    const std::string& joint)
{
    std::string result;
    size_t n = joinable.size();
    if (n == 0)
        return result;
    for (size_t i = 0; i < n - 1; ++i)
        result += joinable[i] + joint;
    result += joinable[n - 1];
    return result;
}

std::string BaseUtils::String::to_lower(std::string text)
{
    boost::to_lower(text);
    return text;
}

bool BaseUtils::String::to_int(const std::string& str, int* result)
{
    const char* first = str.data() + str.find_first_not_of(' ');
    const char* last = str.data() + str.size();
    int _result = 0;
    auto [p, ec] = std::from_chars(first, last, _result);

    if (ec != std::errc())
        return false;

    if (p != last) {
        // not all was consumed. Check whether only space characters left
        const size_t pos = p - str.data();
        const auto hasNonSpaceLeft = str.find_first_not_of(' ', pos) != std::string::npos;
        if (hasNonSpaceLeft)
            return false;
    }

    if (result != nullptr)
        *result = _result;
    return true;
}

std::string BaseUtils::String::trim(const std::string& str, const std::string& whitespace)
{
    const auto strBegin = str.find_first_not_of(whitespace);

    if (strBegin == std::string::npos)
        return "";

    const auto strEnd = str.find_last_not_of(whitespace);
    const auto strRange = strEnd - strBegin + 1;

    return str.substr(strBegin, strRange);
}

std::string BaseUtils::String::trimFront(const std::string& str, const std::string& whitespace)
{
    const auto strBegin = str.find_first_not_of(whitespace);

    if (strBegin == std::string::npos)
        return "";

    return str.substr(strBegin);
}

bool BaseUtils::String::startsWith(const std::string& str, const std::string& substr)
{
    return str.rfind(substr, 0) == 0;
}
