//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      Device/Coord/ICoordSystem.cpp
//! @brief     Implements ICoordSystem class.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "Device/Coord/ICoordSystem.h"
#include "Base/Util/Assert.h"
#include "Device/Data/Datafield.h"
#include <sstream>

ICoordSystem::~ICoordSystem() = default;

std::vector<const IAxis*> ICoordSystem::defaultAxes() const
{
    return convertedAxes(defaultUnits());
}

std::vector<const IAxis*> ICoordSystem::convertedAxes(Coords units) const
{
    std::vector<const IAxis*> result;
    for (size_t i = 0; i < rank(); ++i)
        result.emplace_back(createConvertedAxis(i, units));
    return result;
}

Datafield* ICoordSystem::createConvertedData(const Datafield& data, Coords units) const
{
    return new Datafield(convertedAxes(units), data.flatVector());
}

void ICoordSystem::throwUnitsError(std::string method, std::vector<Coords> available) const
{
    static const std::map<Coords, const char*> axisUnitLabel = {
        {Coords::UNDEFINED, "undefined"}, {Coords::NBINS, "bin"}, {Coords::RADIANS, "rad"},
        {Coords::DEGREES, "deg"},         {Coords::MM, "mm"},     {Coords::QSPACE, "1/nm"},
        {Coords::RQ4, "nm^-4?"}};

    std::stringstream ss;
    ss << "Unit type error in " << method
       << ": unknown or unsupported unit type. Available units "
          "are:\n";
    for (auto unit : available)
        ss << axisUnitLabel.at(unit) << "\n";
    throw std::runtime_error(ss.str());
}

Coords ICoordSystem::substituteDefaultUnits(Coords units) const
{
    return units == Coords::UNDEFINED ? defaultUnits() : units;
}
