//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      Device/Data/Datafield.cpp
//! @brief     Implements template specializations of class Datafield.cpp.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "Device/Data/Datafield.h"
#include "Base/Axis/FixedBinAxis.h"
#include "Base/Axis/Frame.h"
#include "Base/Py/PyCore.h"
#include "Base/Util/Assert.h"
#include "Base/Util/ThreadInfo.h"
#include <algorithm>

Datafield::Datafield(const Frame* frame)
    : m_frame(frame)
{
    ASSERT(m_frame);
    m_values.resize(m_frame->size());
}

Datafield::Datafield(const Frame* frame, const std::vector<double>& values)
    : m_frame(frame)
    , m_values(values)
{
    ASSERT(m_frame);
    ASSERT(m_values.size() == m_frame->size());
}

Datafield::Datafield(const Frame* frame, const std::vector<double>& values,
                     const std::vector<double>& errSigmas)
    : m_frame(frame)
    , m_values(values)
    , m_errSigmas(errSigmas)
{
    ASSERT(m_frame);
    ASSERT(m_values.size() == m_frame->size());
    ASSERT(m_errSigmas.empty() || m_errSigmas.size() == m_values.size());
}

Datafield::Datafield(const std::vector<const IAxis*>& axes)
    : Datafield(new Frame(axes))
{
}

Datafield::Datafield(const std::vector<const IAxis*>& axes, const std::vector<double>& values)
    : Datafield(new Frame(axes), values)
{
}

Datafield::Datafield(const std::vector<const IAxis*>& axes, const std::vector<double>& values,
                     const std::vector<double>& errSigmas)
    : Datafield(new Frame(axes), values, errSigmas)
{
}

Datafield::~Datafield() = default;

Datafield* Datafield::clone() const
{
    auto* data = new Datafield(frame().clonedAxes(), m_values, m_errSigmas);
    return data;
}

void Datafield::setAt(size_t i, double val)
{
    m_values[i] = val;
}

double Datafield::valAt(size_t i) const
{
    return m_values[i];
}

bool Datafield::hasErrorSigmas() const
{
    return !m_errSigmas.empty();
}

const std::vector<double>& Datafield::errorSigmas() const
{
    return m_errSigmas;
}

std::vector<double>& Datafield::errorSigmas()
{
    return m_errSigmas;
}

void Datafield::setAllTo(const double& value)
{
    for (double& v : m_values)
        v = value;
}

void Datafield::setVector(const std::vector<double>& vector)
{
    ASSERT(vector.size() == frame().size());
    m_values = vector;
}

size_t Datafield::rank() const
{
    return frame().rank();
}

size_t Datafield::size() const
{
    ASSERT(frame().size() == m_values.size());
    return frame().size();
}

const Frame& Datafield::frame() const
{
    ASSERT(m_frame);
    return *m_frame;
}

const IAxis& Datafield::axis(size_t k) const
{
    return frame().axis(k);
}

const IAxis& Datafield::xAxis() const
{
    return frame().axis(0);
}

const IAxis& Datafield::yAxis() const
{
    return frame().axis(1);
}


//! Returns true if object have same dimensions
bool Datafield::hasSameSizes(const Datafield& other) const
{
    return frame().hasSameSizes(other.frame());
}

//! Returns true if object have same dimensions and shape of axis
bool Datafield::hasSameShape(const Datafield& other) const
{
    return frame() == other.frame();
}

std::vector<double> Datafield::flatVector() const
{
    std::vector<double> result(size());
    for (size_t i = 0; i < size(); ++i)
        result[i] = (*this)[i];
    return result;
}

void Datafield::scale(double factor)
{
    for (size_t i = 0; i < frame().size(); ++i)
        m_values[i] *= factor;

    if (!m_errSigmas.empty())
        for (size_t i = 0; i < frame().size(); ++i)
            m_errSigmas[i] *= factor;
}

double Datafield::maxVal() const
{
    return *std::max_element(m_values.begin(), m_values.end());
}

double Datafield::minVal() const
{
    return *std::min_element(m_values.begin(), m_values.end());
}

Datafield* Datafield::crop(double xmin, double ymin, double xmax, double ymax) const
{
    const std::unique_ptr<IAxis> xaxis{xAxis().clone()};
    const std::unique_ptr<IAxis> yaxis{yAxis().clone()};
    xaxis->clip(xmin, xmax);
    yaxis->clip(ymin, ymax);

    std::vector<double> out(size());
    size_t iout = 0;
    for (size_t i = 0; i < size(); ++i) {
        double x = frame().projectedCoord(i, 0);
        double y = frame().projectedCoord(i, 1);
        if (xaxis->contains(x) && yaxis->contains(y))
            out[iout++] = m_values[i];
    }
    return new Datafield(frame().clonedAxes(), out);
}

Datafield* Datafield::crop(double xmin, double xmax) const
{
    const std::unique_ptr<IAxis> xaxis{xAxis().clone()};
    xaxis->clip(xmin, xmax);
    std::vector<double> out(size());
    size_t iout = 0;
    for (size_t i = 0; i < size(); ++i) {
        const double x = frame().projectedCoord(i, 0);
        if (xaxis->contains(x))
            out[iout++] = m_values[i];
    }
    return new Datafield(frame().clonedAxes(), out);
}

#ifdef BORNAGAIN_PYTHON

PyObject* Datafield::npArray() const
{
    std::vector<size_t> dimensions;
    for (size_t i = 0; i < rank(); i++)
        dimensions.push_back(axis(i).size());

    // for rot90 of 2-dim arrays to conform with numpy
    if (dimensions.size() == 2)
        std::swap(dimensions[0], dimensions[1]);

    // creating ndarray objects describing size of dimensions
    npy_int ndim_numpy = (int)dimensions.size();
    auto* ndimsizes_numpy = new npy_intp[dimensions.size()];
    for (size_t i = 0; i < dimensions.size(); i++)
        ndimsizes_numpy[i] = dimensions[i];

    // creating standalone numpy array
    PyObject* pyarray = PyArray_SimpleNew(ndim_numpy, ndimsizes_numpy, NPY_DOUBLE);
    delete[] ndimsizes_numpy;
    ASSERT(pyarray);

    // getting pointer to data buffer of numpy array
    double* array_buffer = (double*)PyArray_DATA((PyArrayObject*)pyarray);

    // filling numpy array with output_data
    if (rank() == 2) {
        for (size_t i = 0; i < size(); ++i) {
            std::vector<int> axes_indices = frame().allIndices(i);
            size_t offset =
                axes_indices[0] + axis(0).size() * (axis(1).size() - 1 - axes_indices[1]);
            array_buffer[offset] = (*this)[i];
        }
    } else {
        for (size_t i = 0; i < size(); ++i)
            *array_buffer++ = (*this)[i];
    }

    return pyarray;
}

#endif // BORNAGAIN_PYTHON


Datafield* Datafield::xProjection() const
{
    return create_xProjection(0, static_cast<int>(xAxis().size()) - 1);
}

Datafield* Datafield::xProjection(double yvalue) const
{
    int ybin_selected = static_cast<int>(yAxis().findClosestIndex(yvalue));
    return create_xProjection(ybin_selected, ybin_selected);
}

Datafield* Datafield::xProjection(double ylow, double yup) const
{
    int ybinlow = static_cast<int>(yAxis().findClosestIndex(ylow));
    int ybinup = static_cast<int>(yAxis().findClosestIndex(yup));
    return create_xProjection(ybinlow, ybinup);
}

Datafield* Datafield::yProjection() const
{
    return create_yProjection(0, static_cast<int>(xAxis().size()) - 1);
}

Datafield* Datafield::yProjection(double xvalue) const
{
    int xbin_selected = static_cast<int>(xAxis().findClosestIndex(xvalue));
    return create_yProjection(xbin_selected, xbin_selected);
}

Datafield* Datafield::yProjection(double xlow, double xup) const
{
    int xbinlow = static_cast<int>(xAxis().findClosestIndex(xlow));
    int xbinup = static_cast<int>(xAxis().findClosestIndex(xup));
    return create_yProjection(xbinlow, xbinup);
}

Datafield* Datafield::create_xProjection(int ybinlow, int ybinup) const
{
    std::vector<double> out(xAxis().size());
    for (size_t i = 0; i < size(); ++i) {
        int ybin = static_cast<int>(frame().projectedIndex(i, 1));
        if (ybin >= ybinlow && ybin <= ybinup) {
            double x = frame().projectedCoord(i, 0);
            ASSERT(xAxis().contains(x));
            size_t iout = xAxis().findClosestIndex(x);
            out[iout] += valAt(i);
        }
    }
    return new Datafield({xAxis().clone()}, out);
}

Datafield* Datafield::create_yProjection(int xbinlow, int xbinup) const
{
    std::vector<double> out(yAxis().size());
    for (size_t i = 0; i < size(); ++i) {
        int xbin = static_cast<int>(frame().projectedIndex(i, 0));
        if (xbin >= xbinlow && xbin <= xbinup) {

            // TODO: duplicates code from create_xProjection, move to Frame ?

            double y = frame().projectedCoord(i, 1);
            ASSERT(yAxis().contains(y));
            size_t iout = yAxis().findClosestIndex(y);
            out[iout] += valAt(i);
        }
    }
    return new Datafield({yAxis().clone()}, out);
}
