//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      Device/Histo/SimulationResult.cpp
//! @brief     Implements class SimulationResult.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "Device/Histo/SimulationResult.h"
#include "Base/Axis/IAxis.h"
#include "Base/Util/Assert.h"
#include "Device/Coord/ICoordSystem.h"
#include "Device/Data/Datafield.h"

SimulationResult::SimulationResult() = default;

SimulationResult::SimulationResult(const Datafield& data, const ICoordSystem* coords)
    : m_data(data.clone())
    , m_coordsys(coords)
{
    checkRank();
}

SimulationResult::~SimulationResult() = default;

SimulationResult::SimulationResult(const SimulationResult& other)
    : m_data(other.m_data->clone())
    , m_coordsys(other.m_coordsys->clone())
    , m_title(other.m_title)
{
}

SimulationResult::SimulationResult(SimulationResult&& other) noexcept = default;

SimulationResult& SimulationResult::operator=(SimulationResult&& other) noexcept = default;

size_t SimulationResult::rank() const
{
    return m_data->rank();
}

size_t SimulationResult::size() const
{
    return m_data ? m_data->size() : 0;
}

std::vector<double> SimulationResult::flatVector(Coords units) const
{
    std::unique_ptr<Datafield> f(datafield(units));
    return f->flatVector();
}

Datafield* SimulationResult::datafield(Coords units) const
{
    return new Datafield(m_coordsys->convertedAxes(units), m_data->flatVector());
}

std::pair<double, double> SimulationResult::axisMinMax(size_t i, Coords units) const
{
    return {m_coordsys->calculateMin(i, units), m_coordsys->calculateMax(i, units)};
}

std::string SimulationResult::nameOfAxis(size_t i, Coords units) const
{
    return m_coordsys->nameOfAxis(i, units);
}

const ICoordSystem& SimulationResult::converter() const
{
    return *m_coordsys;
}

double& SimulationResult::operator[](size_t i)
{
    return (*m_data)[i];
}

const double& SimulationResult::operator[](size_t i) const
{
    return (*m_data)[i];
}

#ifdef BORNAGAIN_PYTHON
PyObject* SimulationResult::array(Coords units) const
{
    Datafield data(m_coordsys->convertedAxes(units), m_data->flatVector());
    return data.npArray();
}
#endif

std::vector<double> SimulationResult::convertedBinCenters(Coords units) const
{
    return convertedBinCenters(0, units);
}

std::vector<double> SimulationResult::convertedBinCenters(size_t i_axis, Coords units) const
{
    ASSERT(i_axis < m_coordsys->rank());
    auto* axis = m_coordsys->createConvertedAxis(i_axis, units); // TODO memory leak
    return axis->binCenters();
}

void SimulationResult::checkRank() const
{
    ASSERT(m_data);
    ASSERT(m_coordsys);
    ASSERT(m_data->rank() == m_coordsys->rank());
}
