//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      Device/Histo/SimulationResult.h
//! @brief     Defines class SimulationResult.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#ifndef BORNAGAIN_DEVICE_HISTO_SIMULATIONRESULT_H
#define BORNAGAIN_DEVICE_HISTO_SIMULATIONRESULT_H

#include "Base/Py/PyObject.h" // typedef _object PyObject;
#include "Device/Coord/Tags.h"
#include <memory>
#include <string>
#include <vector>

class Datafield;
class ICoordSystem;

//! Wrapper around Datafield that also provides unit conversions.

class SimulationResult {
public:
    SimulationResult();
    SimulationResult(const Datafield& data, const ICoordSystem* coords);
    ~SimulationResult();

    SimulationResult(const SimulationResult& other);
    SimulationResult(SimulationResult&& other) noexcept;

    SimulationResult& operator=(SimulationResult&& other) noexcept;

    std::vector<double> flatVector(Coords units = Coords::UNDEFINED) const;
    Datafield* datafield(Coords units = Coords::UNDEFINED) const;

    std::pair<double, double> axisMinMax(size_t i, Coords units = Coords::UNDEFINED) const;

    std::string nameOfAxis(size_t i, Coords units = Coords::UNDEFINED) const;

    const double& operator[](size_t i) const;

    const Datafield& data_field() const { return *m_data; } // TODO disambiguate from datafield()

    size_t rank() const;
    size_t size() const;
    bool empty() const { return size() == 0; }

    //! Returns intensity data as Python numpy array
#ifdef BORNAGAIN_PYTHON
    PyObject* array(Coords units = Coords::UNDEFINED) const;
#endif

    std::vector<double> convertedBinCenters(Coords units = Coords::UNDEFINED) const;

    //! Returns axis coordinates as a numpy array. With no parameters given
    //! Returns coordinates of x-axis in default units.
    std::vector<double> convertedBinCenters(size_t i_axis, Coords units = Coords::UNDEFINED) const;

    void setTitle(const std::string& title)
    {
        m_title = title;
    }
    std::string title()
    {
        return m_title;
    }

#ifndef SWIG
    //! Returns underlying unit converter
    const ICoordSystem& converter() const;

    double& operator[](size_t i);

private:
    void checkRank() const;

    std::unique_ptr<Datafield> m_data;
    std::unique_ptr<const ICoordSystem> m_coordsys;

    std::string m_title;
#endif // SWIG
};

#endif // BORNAGAIN_DEVICE_HISTO_SIMULATIONRESULT_H
