//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      Device/Pol/PolFilter.cpp
//! @brief     Implements class DetectionProperties.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "Device/Pol/PolFilter.h"
#include "Base/Spin/SpinMatrix.h"

PolFilter::PolFilter(R3 direction, double efficiency, double total_transmission)
{
    double aplus = total_transmission * (1.0 + efficiency);
    double amin = total_transmission * (1.0 - efficiency);
    if (aplus < 0.0 || aplus > 1.0 || amin < 0.0 || amin > 1.0)
        throw std::runtime_error("Invalid efficiency and transmission for polarization analyzer");
    if (efficiency == 0.0 || total_transmission == 0.0 || direction.mag() == 0.0) {
        m_direction = {};
        m_efficiency = 0;
        m_total_transmission = total_transmission;
    } else {
        m_direction = direction.unit_or_throw();
        m_efficiency = efficiency;
        m_total_transmission = total_transmission;
    }
}

PolFilter::PolFilter()
    : PolFilter({}, 0.0, 1.0)
{
}

SpinMatrix PolFilter::matrix() const
{
    if (m_direction.mag() == 0.0 || m_efficiency == 0.0)
        return m_total_transmission * SpinMatrix::One();
    R3 v = m_direction.unit_or_throw() * m_efficiency;
    return 2 * m_total_transmission * SpinMatrix::FromBlochVector(v);
}

R3 PolFilter::analyzerDirection() const
{
    return m_direction;
}

double PolFilter::analyzerEfficiency() const
{
    return m_efficiency;
}

double PolFilter::totalTransmission() const
{
    return m_total_transmission;
}
