"""
Functional test: Magnetic cylinders in DWBA with zero magnetic field
"""

import bornagain as ba
from bornagain import angstrom, ba_plot as bp, deg, R3, nm


def get_sample():
    # defining materials
    mAmbience = ba.RefractiveMaterial("Vacuum", 0, 0)
    mSubstrate = ba.RefractiveMaterial("Substrate", 15e-6, 0)

    magnetization = R3(0, 1e6, 0)

    magParticle = ba.RefractiveMaterial("magParticle", 5e-6, 0,
                                        magnetization)
    # collection of particles
    cylinder_ff = ba.Cylinder(5*nm, 5*nm)
    cylinder = ba.Particle(magParticle, cylinder_ff)

    particle_layout = ba.ParticleLayout()
    particle_layout.addParticle(cylinder, 1)
    interference = ba.InterferenceNone()
    particle_layout.setInterference(interference)

    # vacuum layer with particles and substrate form multi layer
    vacuum_layer = ba.Layer(mAmbience)
    vacuum_layer.addLayout(particle_layout)
    substrate_layer = ba.Layer(mSubstrate, 0)
    sample = ba.MultiLayer()
    sample.addLayer(vacuum_layer)
    sample.addLayer(substrate_layer)

    return sample


def get_simulation(sample, pol_dir, efficiency):
    z_up = R3(0, 0, 1)
    n = bp.simargs['n']
    beam = ba.Beam(1e9, 0.1*nm, 0.2*deg)
    detector = ba.SphericalDetector(n, -1*deg, 1*deg, n, 0, 2*deg)

    beam.setPolarization(pol_dir)
    detector.setAnalyzer(z_up, efficiency, 0.5)

    return ba.ScatteringSimulation(beam, sample, detector)


def simulate(title, pol_dir, efficiency):
    sample = get_sample()
    simulation = get_simulation(sample, pol_dir, efficiency)
    result = simulation.simulate()
    result.setTitle(title)
    return result


if __name__ == '__main__':
    bp.parse_args(sim_n=100, fontsize=12)
    z_up = R3(0, 0, 1)
    z_dn = R3(0, 0, -1)
    results = []
    results.append(simulate('++', z_up, 1))
    results.append(simulate('+-', z_dn, 1))
    results.append(simulate('-+', z_up, 1))
    results.append(simulate('--', z_dn, 1))
    bp.make_plot(results, 2)
