#!/usr/bin/env python3
"""
Simulation demo: magnetic spheres in substrate
"""
import bornagain as ba
from bornagain import ba_plot as bp, deg, nm, R3


def get_sample():
    """
    Returns a sample with magnetic spheres in the substrate.
    """

    # Define materials
    magnetic_field = R3(0, 0, 1e7)
    material_Particle = ba.RefractiveMaterial("Particle", 2e-05, 4e-07,
                                              magnetic_field)
    material_Substrate = ba.RefractiveMaterial("Substrate", 7e-06, 1.8e-07)
    material_Vacuum = ba.RefractiveMaterial("Vacuum", 0, 0)

    # Define form factors
    ff = ba.Sphere(5*nm)

    # Define particles
    particle = ba.Particle(material_Particle, ff)
    particle_position = R3(0, 0, -10*nm)
    particle.translate(particle_position)

    # Define particle layouts
    layout = ba.ParticleLayout()
    layout.addParticle(particle, 1)
    layout.setTotalParticleSurfaceDensity(0.01)

    # Define layers
    layer_1 = ba.Layer(material_Vacuum)
    layer_2 = ba.Layer(material_Substrate)
    layer_2.addLayout(layout)

    # Define sample
    sample = ba.MultiLayer()
    sample.addLayer(layer_1)
    sample.addLayer(layer_2)

    return sample


def get_simulation(sample):
    beam = ba.Beam(1e12, 0.1*nm, 0.5*deg)
    detector = ba.SphericalDetector(bp.simargs['n'], 6*deg, 0, 3*deg)

    polarizer_dir = R3(0, 0, 1)
    analyzer_dir = R3(0, 0, -1)
    beam.setPolarization(polarizer_dir)
    detector.setAnalyzer(analyzer_dir, 1, 0.5)

    return ba.ScatteringSimulation(beam, sample, detector)


if __name__ == '__main__':
    bp.parse_args(sim_n=200)
    sample = get_sample()
    simulation = get_simulation(sample)
    result = simulation.simulate()
    bp.plot_simulation_result(result)
