#!/usr/bin/env python3
"""
"""
import bornagain as ba
from bornagain import ba_plot as bp, deg, nm
from matplotlib import pyplot as plt


def get_sample(hasVariance, xi):
    # Define materials
    m_air = ba.RefractiveMaterial("Air", 0, 0)
    m_particle = ba.RefractiveMaterial("Particle", 0.0006, 2e-08)
    m_substrate = ba.RefractiveMaterial("Substrate", 6e-06, 2e-08)

    # Define particles
    R = 2.5*nm
    ff = ba.Spheroid(R, R)
    particle = ba.Particle(m_particle, ff)

    # Define interference function
    lattice = ba.SquareLattice2D(10*nm, xi)
    interference = ba.Interference2DLattice(lattice)
    interference_pdf = ba.Profile2DCauchy(500*nm, 500*nm, 0)
    interference.setDecayFunction(interference_pdf)
    if hasVariance:
        interference.setPositionVariance(0.3*nm)

    # Define particle layout
    layout = ba.ParticleLayout()
    layout.addParticle(particle)
    layout.setInterference(interference)

    # Define layers
    l_air = ba.Layer(m_air)
    l_air.addLayout(layout)
    l_substrate = ba.Layer(m_substrate)

    # Define sample
    sample = ba.MultiLayer()
    sample.addLayer(l_air)
    sample.addLayer(l_substrate)
    return sample


def get_simulation(sample):
    n = bp.simargs['n']
    beam = ba.Beam(1e8, 0.1*nm, 0.2*deg)
    det = ba.SphericalDetector(n, -2*deg, 2*deg, n, 0, 3*deg)
    return ba.ScatteringSimulation(beam, sample, det)


def run_one(hasVariance, xi, nPlot, title):
    sample = get_sample(hasVariance, xi)
    simulation = get_simulation(sample)
    result = simulation.simulate()

    plt.subplot(3, 2, nPlot)

    return bp.plot_simres(result,
                          title=title,
                          intensity_max=3e7,
                          intensity_min=3e0,
                          zlabel=None,
                          aspect='equal',
                          with_cb=False)


if __name__ == '__main__':
    bp.parse_args(sim_n=200)

    fig, axs = plt.subplots(3, 2, figsize=(10, 13))

    xi1 = 5*deg
    xi2 = 15*deg
    im = run_one(False, 0*deg, 1, r"$\xi=0^\circ$, fixed positions")
    run_one(True, 0*deg, 2, r"position variance 0.3 nm")
    run_one(False, xi1, 3, r"$\xi=5^\circ$, fixed positions")
    run_one(True, xi1, 4, r"position variance 0.3 nm")
    run_one(False, xi2, 5, r"$\xi=15^\circ$, fixed positions")
    run_one(True, xi2, 6, r"position variance 0.3 nm")

    plt.subplots_adjust(bottom=0.05, left=0.1)
    plt.colorbar(im, cax=plt.axes([0.93, 0.36, 0.03, 0.21]))

    bp.show_or_export()
