#!/usr/bin/env python3
"""
An example that shows how to simulate
a non-perfect polarizer and analyzer
following Devishvili et al., Rev. Sci. Instrum. 84, 025112 (2013)
"""
import numpy
import matplotlib.pyplot as plt

import bornagain as ba
from bornagain import angstrom, ba_plot as bp, deg, R3

sldFe = (8.0241e-06, 6.0448e-10)
sldPd = (4.0099e-6, 1.3019e-09)
sldMgO = (5.9803e-06, 9.3996e-12)

magnetizationMagnitude = 1.6e6
angle = 0
magnetizationVector = R3(magnetizationMagnitude*numpy.sin(angle*deg),
                            magnetizationMagnitude*numpy.cos(angle*deg), 0)


def get_sample(*, magnetization=magnetizationVector):
    """
    Define sample and returns it
    """

    # create materials
    mat_vacuum = ba.MaterialBySLD("Vacuum", 0, 0)
    mat_Pd = ba.MaterialBySLD("Pd", *sldPd)
    mat_Fe = ba.MaterialBySLD("Fe", *sldFe, magnetizationVector)
    mat_substrate = ba.MaterialBySLD("MgO", *sldMgO)

    # create layers
    layer_vacuum = ba.Layer(mat_vacuum)
    layer_Pd = ba.Layer(mat_Pd, 120*angstrom)
    layer_Fe = ba.Layer(mat_Fe, 1000*angstrom)
    layer_substrate = ba.Layer(mat_substrate)

    roughness = ba.LayerRoughness(20*angstrom)

    # create sample
    sample = ba.MultiLayer()

    sample.addLayer(layer_vacuum)
    sample.addLayerWithTopRoughness(layer_Pd, roughness)
    sample.addLayerWithTopRoughness(layer_Fe, roughness)
    sample.addLayerWithTopRoughness(layer_substrate, roughness)

    return sample


def run_simulation(*,
                   polarizer_dir=R3(0, 1, 0),
                   polarizer_efficiency=1,
                   analyzer_dir=R3(0, 1, 0),
                   analyzer_efficiency=1):
    """
    Runs simulation and returns its result.
    """
    sample = get_sample()

    qzs = numpy.linspace(0.1, 1.5, 1500)
    distr = ba.DistributionGaussian(0., 1., 25, 4.)

    scan = ba.QzScan(qzs)
    scan.setAbsoluteQResolution(distr, 0.008)

    scan.setPolarization(polarizer_dir*polarizer_efficiency)
    scan.setAnalyzer(analyzer_dir, analyzer_efficiency, 0.5)

    simulation = ba.SpecularSimulation(scan, sample)
    simulation.setBackground(ba.ConstantBackground(1e-7))

    return simulation.simulate()


def plot(data, labels):

    plt.figure()
    plt.yscale('log')
    for d, l in zip(data, labels):
        plt.plot(numpy.array(d.convertedBinCenters(ba.Coords_QSPACE)),
                 d.array(ba.Coords_QSPACE),
                 label=l,
                 linewidth=1)

    plt.legend(loc='upper right')

    bp.inside_ticks()

    plt.xlabel("$Q$ [nm${}^{-1}$]")
    plt.ylabel("Reflectivity")

    plt.tight_layout()


if __name__ == '__main__':
    bp.parse_args()

    polarizer_efficiency = 0.986
    analyzer_efficiency = 0.970

    results_pp = run_simulation(polarizer_dir=R3(0, 1, 0),
                                analyzer_dir=R3(0, 1, 0),
                                polarizer_efficiency=polarizer_efficiency,
                                analyzer_efficiency=analyzer_efficiency)
    results_mm = run_simulation(polarizer_dir=R3(0, -1, 0),
                                analyzer_dir=R3(0, -1, 0),
                                polarizer_efficiency=polarizer_efficiency,
                                analyzer_efficiency=analyzer_efficiency)

    results_pm = run_simulation(polarizer_dir=R3(0, 1, 0),
                                analyzer_dir=R3(0, -1, 0),
                                polarizer_efficiency=polarizer_efficiency,
                                analyzer_efficiency=analyzer_efficiency)
    results_mp = run_simulation(polarizer_dir=R3(0, -1, 0),
                                analyzer_dir=R3(0, 1, 0),
                                polarizer_efficiency=polarizer_efficiency,
                                analyzer_efficiency=analyzer_efficiency)

    plot([results_pp, results_mm, results_pm, results_mp],
         ["$++$", "$--$", "$+-$", "$-+$"])

    bp.show_or_export()
