#!/usr/bin/env python3
"""
An example of computing splin-flip reflectivity from
a magnetized sample.
"""
from math import sqrt
import bornagain as ba
from bornagain import angstrom, ba_plot as bp, deg, nm, R3
import matplotlib.pyplot as plt


def get_sample():
    """
    Defines sample and returns it
    """

    # Define materials
    material_Ambient = ba.MaterialBySLD("Ambient", 0, 0)
    h = 1e8
    magnetic_field = R3(1/2*h, sqrt(3)/2*h, 0)
    material_Layer = ba.MaterialBySLD("Layer", 0.0001, 1e-08,
                                      magnetic_field)
    material_Substrate = ba.MaterialBySLD("Substrate", 7e-05, 2e-06)

    # Define layers
    layer_1 = ba.Layer(material_Ambient)
    layer_2 = ba.Layer(material_Layer, 10*nm)
    layer_3 = ba.Layer(material_Substrate)

    # Define sample
    sample = ba.MultiLayer()
    sample.addLayer(layer_1)
    sample.addLayer(layer_2)
    sample.addLayer(layer_3)

    return sample


def run_simulation(sample,
                   polarizer_dir=R3(0, 1, 0),
                   analyzer_dir=R3(0, 1, 0)):
    """
    Runs simulation and returns its result.
    """
    scan = ba.AlphaScan(500, 5*deg/500, 5*deg)
    scan.setWavelength(1.54*angstrom)
    scan.setPolarization(polarizer_dir)
    scan.setAnalyzer(analyzer_dir, 1, 0.5)

    simulation = ba.SpecularSimulation(scan, sample)

    return simulation.simulate()


def plot(data, labels):

    plt.figure()
    plt.yscale('log')
    for d, l in zip(data, labels):
        plt.plot(d.convertedBinCenters(), d.array(), label=l, linewidth=1)

    plt.legend(loc='upper right')

    bp.inside_ticks()

    plt.xlabel(r"$\alpha_{\rm i} \;(^\circ)$")
    plt.ylabel("Reflectivity")

    plt.tight_layout()


if __name__ == '__main__':
    bp.parse_args()

    sample = get_sample()

    results_pp = run_simulation(sample, R3(0, 1, 0), R3(0, 1, 0))
    results_mm = run_simulation(sample, R3(0, -1, 0), R3(0, -1, 0))
    results_pm = run_simulation(sample, R3(0, 1, 0), R3(0, -1, 0))
    results_mp = run_simulation(sample, R3(0, -1, 0), R3(0, 1, 0))

    plot([results_pp, results_mm, results_pm, results_mp],
         ["$++$", "$--$", "$+-$", "$-+$"])

    bp.show_or_export()
