//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/Model/Data/JobItemUtils.cpp
//! @brief     Implements class JobItemUtils
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/Model/Data/JobItemUtils.h"
#include "Device/Data/Datafield.h"
#include "Device/Histo/SimulationResult.h"
#include "GUI/Model/Data/DataItem.h"
#include "GUI/Model/Device/InstrumentItems.h"
#include "GUI/Support/Util/CoordName.h"

namespace {

//! Updates axes' titles
void updateAxesTitle(DataItem* intensityItem, const ICoordSystem& converter, Coords units)
{
    intensityItem->setXaxisTitle(QString::fromStdString(converter.nameOfAxis(0, units)));
    if (converter.rank() > 1)
        intensityItem->setYaxisTitle(QString::fromStdString(converter.nameOfAxis(1, units)));
}

} // namespace


//! Updates axes of Datafield in IntensityData item to correspond with axes units selection.
//! InstrumentItem is used to get domain's detector map for given units.

void GUI::Model::JobItemUtils::updateDataAxes(DataItem* dataItem,
                                              const InstrumentItem* instrumentItem)
{
    ASSERT(dataItem);

    if (!instrumentItem) {
        // special case when reading old project files: project failed on load instrument
        // but we want to try to recover the rest.
        return;
    }

    Datafield* oldData = dataItem->datafield();
    if (!oldData)
        return;

    Coords axes_units = GUI::Util::CoordName::coordFromName(dataItem->currentAxesUnits());

    auto* const converter = instrumentItem->createCoordSystem();

    auto newData = std::make_unique<Datafield>(converter->convertedAxes(axes_units),
                                               oldData->flatVector(), oldData->errorSigmas());

    dataItem->setDatafield(newData.release());
    dataItem->setAxesRangeToData();
    updateAxesTitle(dataItem, *converter, axes_units);
}

//! Sets axes units suitable for given instrument.

void GUI::Model::JobItemUtils::setDataItemCoords(DataItem* dataItem,
                                                 const InstrumentItem* instrumentItem)
{
    ASSERT(instrumentItem);
    auto* const converter = instrumentItem->createCoordSystem();
    if (!converter)
        return;
    setDataItemCoords(dataItem, *converter);
}

void GUI::Model::JobItemUtils::setDataItemCoords(DataItem* dataItem, const ICoordSystem& converter)
{
    ComboProperty combo = availableUnits(converter);
    dataItem->setAxesUnitsCombo(combo);
}

void GUI::Model::JobItemUtils::createDefaultDetectorMap(DataItem* dataItem,
                                                        const InstrumentItem* instrumentItem)
{
    auto* const converter = instrumentItem->createCoordSystem();
    auto output_data = std::make_unique<Datafield>(converter->defaultAxes());
    dataItem->setDatafield(output_data.release());
    setDataItemCoords(dataItem, *converter);
    updateAxesTitle(dataItem, *converter, converter->defaultUnits());
}

void GUI::Model::JobItemUtils::setResults(DataItem* dataItem, const SimulationResult& result)
{
    if (dataItem->datafield() == nullptr) {
        const auto& converter = result.converter();
        GUI::Model::JobItemUtils::setDataItemCoords(dataItem, converter);
        updateAxesTitle(dataItem, converter, converter.defaultUnits());
    }
    auto selected_units = GUI::Util::CoordName::coordFromName(dataItem->currentAxesUnits());
    dataItem->setDatafield(result.datafield(selected_units));
}

ComboProperty GUI::Model::JobItemUtils::availableUnits(const ICoordSystem& converter)
{
    ComboProperty result;
    for (auto units : converter.availableUnits()) {
        auto unit_name = GUI::Util::CoordName::nameFromCoord(units);
        if (!unit_name.isEmpty())
            result << unit_name;
    }

    result.setCurrentValue(GUI::Util::CoordName::nameFromCoord(converter.defaultUnits()));
    return result;
}
