//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/Support/IO/DataLoaders1D.h
//! @brief     Defines class DataLoaders1D
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2021
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#ifndef BORNAGAIN_GUI_SUPPORT_IO_DATALOADERS1D_H
#define BORNAGAIN_GUI_SUPPORT_IO_DATALOADERS1D_H

#include <QVector>

#include <functional>
#include <map>

class AbstractDataLoader1D;

//! Collection of all available data loaders for 1D files

class DataLoaders1D {
public:
    ~DataLoaders1D();

    //! The one and only instance
    static DataLoaders1D& instance();

    //! Register a built-in loader with the given class name and factory function
    void addBuiltInLoader(const QString& persistentClassName,
                          const std::function<AbstractDataLoader1D*()>& create);

    //! add a user defined loader
    void addUserDefinedLoader(AbstractDataLoader1D* loader);

    //! all defined loaders. A null element in the list defines a separator
    //! The returned pointers are the same over the lifetime of the DataLoaders instance, therefore
    //! they can be used for comparison.
    QVector<AbstractDataLoader1D*> loaders() const;

    //! The last 10 recently used loaders.
    QVector<AbstractDataLoader1D*> recentlyUsedLoaders() const;

    //! Notify loader was recently used
    // void setRecentlyUsedLoader(const AbstractDataLoader1D* loader);

    //! Create loader from the given persistent name
    AbstractDataLoader1D* createFromPersistentName(const QString& persistentClassName);

private:
    DataLoaders1D() = default;

    //! create all default built in loaders
    void initBuiltInLoaders();

private:
    QVector<AbstractDataLoader1D*> m_builtInLoaders;
    QVector<AbstractDataLoader1D*> m_recentlyUsedLoaders;
    QVector<AbstractDataLoader1D*> m_userDefinedLoaders;
    std::map<QString, std::function<AbstractDataLoader1D*()>> m_createLoaders;
};

#endif // BORNAGAIN_GUI_SUPPORT_IO_DATALOADERS1D_H
