//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/Support/IO/ImportDataInfo.cpp
//! @brief     Implements ImportDataInfo helper struct
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/Support/IO/ImportDataInfo.h"
#include "Base/Axis/Frame.h"
#include "Base/Axis/IAxis.h"
#include "Device/Coord/CoordSystem1D.h"
#include "Device/Data/Datafield.h"
#include "GUI/Support/Util/CoordName.h"
#include "GUI/Support/Util/Error.h"
#include <memory>

ImportDataInfo::ImportDataInfo(ImportDataInfo&& other)
    : m_data(std::move(other.m_data))
    , m_coords(other.m_coords)
{
}

ImportDataInfo::ImportDataInfo(std::unique_ptr<Datafield> data, Coords units)
    : m_data(std::move(data))
    , m_coords(units)
{
    checkValidity();
}

ImportDataInfo::ImportDataInfo(std::unique_ptr<Datafield> data, const QString& units_label)
    : m_data(std::move(data))
    , m_coords(GUI::Util::CoordName::coordFromName(units_label))
{
    checkValidity();
}

ImportDataInfo::ImportDataInfo(Datafield&& data, Coords units)
    : ImportDataInfo(std::make_unique<Datafield>(std::move(data)), units)
{
}

ImportDataInfo::~ImportDataInfo() = default;

ImportDataInfo::operator bool() const
{
    return static_cast<bool>(m_data);
}

std::unique_ptr<Datafield> ImportDataInfo::clonedData() const&
{
    if (!m_data)
        return nullptr;
    return std::unique_ptr<Datafield>(m_data->clone());
}

std::unique_ptr<Datafield> ImportDataInfo::intensityData() &&
{
    return std::move(m_data);
}

size_t ImportDataInfo::dataRank() const
{
    if (!m_data)
        return 0;
    return m_data->rank();
}

QString ImportDataInfo::unitsLabel() const
{
    return GUI::Util::CoordName::nameFromCoord(m_coords);
}

QString ImportDataInfo::axisLabel(size_t axis_index) const
{
    if (!m_data)
        return "";

    const size_t rank = m_data->rank();
    if (rank == 2)
        return axis_index == 0 ? "X [nbins]" : "Y [nbins]";
    if (rank == 1) {
        if (axis_index > 0)
            return "Signal [a.u.]";
        return QString::fromStdString(AngularReflectometryCoords::nameOfAxis0(m_coords));
    }
    throw Error("Error in ImportDataInfo::axisLabel: unsupported data type");
}

void ImportDataInfo::checkValidity()
{
    static std::map<size_t, std::vector<Coords>> available_units = {
        {1u, AngularReflectometryCoords::availableUnits0()}, {2u, {Coords::NBINS}}};

    if (!m_data)
        return;
    auto iter = available_units.find(m_data->rank());
    if (iter == available_units.end())
        throw Error("Error in ImportDataInfo constructor: unsupported data type");
    for (auto& value : iter->second)
        if (m_coords == value)
            return;

    throw Error("Error in ImportDataInfo constructor: inacceptable units passed.");
}
