//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/View/Instrument/PolarizationAnalysisEditor.cpp
//! @brief     Implements class PolarizationAnalysisEditor
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/View/Instrument/PolarizationAnalysisEditor.h"
#include "GUI/Model/Device/InstrumentItems.h"
#include "GUI/View/Numeric/DoubleSpinBox.h"
#include "GUI/View/Numeric/NumWidgetUtils.h"
#include "GUI/View/Tool/GroupBoxCollapser.h"
#include <QLabel>

PolarizationAnalysisEditor::PolarizationAnalysisEditor(QWidget* parent, InstrumentItem* instrument)
    : QGroupBox("Polarization analysis", parent)
    , m_instrument(instrument)
{
    ASSERT(instrument);

    auto* layout = new QVBoxLayout(this);

    auto* checkbox = new QCheckBox(QString("Enable %1").arg(title()));
    checkbox->setChecked(m_instrument->withPolarizerAnalyzer());
    layout->addWidget(checkbox);

    auto* polarizerAnalyzerWidget = new QWidget(this);
    auto* formlayout = new QFormLayout(polarizerAnalyzerWidget);
    formlayout->setFieldGrowthPolicy(QFormLayout::FieldsStayAtSizeHint);

    addVector(formlayout, m_instrument->polarization());
    addVector(formlayout, m_instrument->analyzerDirection());
    formlayout->addRow(GUI::Util::labelWithUnit(m_instrument->analyzerEfficiency()) + ":",
                       createSpinBox(m_instrument->analyzerEfficiency()));
    formlayout->addRow(GUI::Util::labelWithUnit(m_instrument->analyzerTotalTransmission()) + ":",
                       createSpinBox(m_instrument->analyzerTotalTransmission()));

    layout->addWidget(polarizerAnalyzerWidget);
    polarizerAnalyzerWidget->setVisible(m_instrument->withPolarizerAnalyzer());

    connect(checkbox, &QCheckBox::toggled, [=](bool b) {
        polarizerAnalyzerWidget->setVisible(b);
        m_instrument->setWithPolarizerAnalyzer(b);
        emit dataChanged();
    });

    auto* collapser = GroupBoxCollapser::installIntoGroupBox(this);
    collapser->setExpanded(instrument->isExpandPolarizerAlanyzer());
    connect(collapser, &GroupBoxCollapser::toggled, this,
            [instrument](bool b) { instrument->setExpandPolarizerAlanyzer(b); });
}

DoubleSpinBox* PolarizationAnalysisEditor::createSpinBox(DoubleProperty& d)
{
    auto* sb = new DoubleSpinBox(d);
    connect(sb, &DoubleSpinBox::baseValueChanged, [this, &d](double v) {
        if (d.value() != v) {
            d.setValue(v);
            emit dataChanged();
        }
    });
    return sb;
}

void PolarizationAnalysisEditor::addVector(QFormLayout* parentLayout, VectorProperty& d)
{
    auto* layout = new QHBoxLayout;

    const auto add = [this, layout](DoubleProperty& d) {
        layout->addWidget(new QLabel(GUI::Util::labelWithUnit(d) + ":"));
        layout->addWidget(createSpinBox(d));
    };

    add(d.x());
    add(d.y());
    add(d.z());

    layout->addItem(new QSpacerItem(0, 0, QSizePolicy::Expanding));

    parentLayout->addRow(d.label() + ":", layout);
}
