//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/View/PlotSpecular/SpecularDataPropertyWidget.cpp
//! @brief     Implements class SpecularDataPropertyWidget
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/View/PlotSpecular/SpecularDataPropertyWidget.h"
#include "GUI/Model/Data/ComboProperty.h"
#include "GUI/Model/Data/SpecularDataItem.h"
#include "GUI/Model/Device/AxisItems.h"
#include "GUI/Model/Project/ProjectDocument.h"
#include "GUI/View/Numeric/DoubleSpinBox.h"
#include "GUI/View/Numeric/NumWidgetUtils.h"
#include "GUI/View/Tool/LayoutUtils.h"
#include <QAction>
#include <QCheckBox>
#include <QFormLayout>
#include <QLineEdit>

SpecularDataPropertyWidget::SpecularDataPropertyWidget(QWidget* parent)
    : DataPropertyWidget(parent)
{
}

void SpecularDataPropertyWidget::createPanelElements()
{
    if (allSpecularDataItems().size() == 0)
        return;

    for (auto item : allSpecularDataItems())
        disconnect(item, nullptr, this, nullptr);

    GUI::Util::Layout::clearLayout(m_mainLayout);
    m_updaters.clear();

    m_mainLayout->addRow(
        "Axes units:",
        GUI::Util::createComboBox([=] { return currentSpecularDataItem()->axesUnitsCombo(); },
                                  [=](const QString& newVal) {
                                      for (auto item : allSpecularDataItems())
                                          item->setCurrentAxesUnits(newVal);
                                      emit axesRangeResetRequested();
                                      gProjectDocument.value()->setModified();
                                  },
                                  &m_updaters));

    // -- x-axis
    auto* xGroup = new QGroupBox("X axis", this);
    auto* xFormLayout = new QFormLayout(xGroup);
    xFormLayout->setContentsMargins(0, 0, 0, 0);
    xFormLayout->setSpacing(5);

    xFormLayout->addRow("Min:", GUI::Util::createDoubleSpinbox(
                                    [=] { return currentSpecularDataItem()->xAxisItem()->min(); },
                                    [=](double newValue) {
                                        for (auto item : allSpecularDataItems())
                                            item->xAxisItem()->setMin(newValue);
                                        gProjectDocument.value()->setModified();
                                    },
                                    &m_updaters));

    xFormLayout->addRow("Max:", GUI::Util::createDoubleSpinbox(
                                    [=] { return currentSpecularDataItem()->xAxisItem()->max(); },
                                    [=](double newValue) {
                                        for (auto item : allSpecularDataItems())
                                            item->xAxisItem()->setMax(newValue);
                                        gProjectDocument.value()->setModified();
                                    },
                                    &m_updaters));

    xFormLayout->addRow(
        "Title:",
        GUI::Util::createTextEdit([=] { return currentSpecularDataItem()->xAxisItem()->title(); },
                                  [=](QString newText) {
                                      for (auto item : allSpecularDataItems())
                                          item->xAxisItem()->setTitle(newText);
                                      gProjectDocument.value()->setModified();
                                  },
                                  &m_updaters));

    m_mainLayout->addRow(xGroup);

    // -- y-axis
    auto* yGroup = new QGroupBox("Y axis", this);
    auto* yFormLayout = new QFormLayout(yGroup);
    yFormLayout->setContentsMargins(0, 0, 0, 0);
    yFormLayout->setSpacing(5);

    yFormLayout->addRow("Min:", GUI::Util::createDoubleSpinbox(
                                    [=] { return currentSpecularDataItem()->yAxisItem()->min(); },
                                    [=](double newValue) {
                                        for (auto item : mainSpecularDataItems())
                                            item->yAxisItem()->setMin(newValue);
                                        gProjectDocument.value()->setModified();
                                    },
                                    &m_updaters));

    yFormLayout->addRow("Max:", GUI::Util::createDoubleSpinbox(
                                    [=] { return currentSpecularDataItem()->yAxisItem()->max(); },
                                    [=](double newValue) {
                                        for (auto item : mainSpecularDataItems())
                                            item->yAxisItem()->setMax(newValue);
                                        gProjectDocument.value()->setModified();
                                    },
                                    &m_updaters));

    yFormLayout->addRow(
        "Title:",
        GUI::Util::createTextEdit([=] { return currentSpecularDataItem()->yAxisItem()->title(); },
                                  [=](QString newText) {
                                      for (auto item : mainSpecularDataItems())
                                          item->yAxisItem()->setTitle(newText);
                                      gProjectDocument.value()->setModified();
                                  },
                                  &m_updaters));

    yFormLayout->addRow(GUI::Util::createCheckBox(
        "log10", [=] { return currentSpecularDataItem()->yAxisItem()->isLogScale(); },
        [=](bool b) {
            for (auto item : allSpecularDataItems())
                item->yAxisItem()->setLogScale(b);
            gProjectDocument.value()->setModified();
        },
        &m_updaters));

    m_mainLayout->addRow(yGroup);

    updateUIValues();

    // react on external changes (e.g. zooming in customplot shall update the axis values)
    connect(currentSpecularDataItem(), &DataItem::itemAxesRangeChanged, this,
            &SpecularDataPropertyWidget::updateUIValues, Qt::UniqueConnection);

    // update coordinates on axes units change
    for (auto item : allSpecularDataItems())
        connect(item, &DataItem::axesUnitsChanged, this,
                &SpecularDataPropertyWidget::updateItemCoords, Qt::UniqueConnection);
}
