//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      Resample/Element/DepthprobeElement.h
//! @brief     Defines class DepthprobeElement.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#ifdef SWIG
#error no need to expose this header to Swig
#endif // SWIG
#ifndef BORNAGAIN_RESAMPLE_ELEMENT_DEPTHPROBEELEMENT_H
#define BORNAGAIN_RESAMPLE_ELEMENT_DEPTHPROBEELEMENT_H

#include "Resample/Element/IElement.h"
#include <heinz/Vectors3D.h>
#include <valarray>

class IAxis;

class DepthprobeElement : public IElement {
public:
    DepthprobeElement(double wavelength, double alpha_i, const IAxis* z_positions,
                      bool calculation_flag);
    DepthprobeElement(const DepthprobeElement& other) = delete;
    DepthprobeElement(DepthprobeElement&& other) noexcept;

    ~DepthprobeElement();

    double wavelength() const { return m_wavelength; }
    double alphaI() const { return m_alpha_i; }
    R3 getKi() const;

    template <typename T>
    void setIntensities(T&& intensities)
    {
        static_assert(
            std::is_assignable<std::valarray<double>, typename std::decay<T>::type>::value,
            "Error in DepthprobeElement::setIntensities: wrong type of input data.");
        m_intensities = std::forward<T>(intensities);
    }

    const std::valarray<double>& getIntensities() const { return m_intensities; }

    void setZPositions(const IAxis* z_positions) { m_z_positions = z_positions; }
    const IAxis* getZPositions() const { return m_z_positions; }

    size_t size() const { return m_intensities.size(); }

    //! if it's false, zero intensity is assigned to the element
    bool isCalculated() const { return m_calculation_flag; }

private:
    const double m_wavelength;
    const double m_alpha_i;              //!< incident angle of the beam
    std::valarray<double> m_intensities; //!< simulated intensity for the set of z positions
    const IAxis* m_z_positions;          //!< positions (lower z corresponds to a greater depth)
    const bool m_calculation_flag;
};

#endif // BORNAGAIN_RESAMPLE_ELEMENT_DEPTHPROBEELEMENT_H
