//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      Sim/Computation/SpecularComputation.cpp
//! @brief     Implements class SpecularComputation.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "Sim/Computation/SpecularComputation.h"
#include "Base/Progress/ProgressHandler.h"
#include "Resample/Element/SpecularElement.h"
#include "Resample/Flux/IFlux.h"
#include "Resample/Processed/ReSample.h"
#include "Resample/Specular/ComputeFluxMagnetic.h"
#include "Resample/Specular/ComputeFluxScalar.h"

SpecularComputation::SpecularComputation(const ReSample& re_sample,
                                         const SimulationOptions& options,
                                         ProgressHandler& progress)
    : IComputation(std::move(re_sample), options, progress)
{
}

SpecularComputation::~SpecularComputation() = default;

//! Performs the actual simulation, and sets intensity, for given range of SpecularElement%s.
double SpecularComputation::simulatedIntensity(SpecularElement& ele)
{
    if (!ele.isCalculated())
        return 0;

    const SliceStack& slices = m_re_sample.averageSlices();
    std::vector<complex_t> kz_stack = ele.produceKz(slices);

    double refl;

    if (m_re_sample.polarizing()) {
        const SpinMatrix R = Compute::SpecularMagnetic::topLayerR(slices, kz_stack, true);

        const SpinMatrix& polMatrix = ele.polarizer();
        const SpinMatrix& anaMatrix = ele.analyzer();

        const complex_t trace = (polMatrix * R.adjoint() * anaMatrix * R).trace();

        refl = std::abs(trace);

    } else {
        const complex_t R = Compute::SpecularScalar::topLayerR(slices, kz_stack);

        refl = std::norm(R);
    }

    stepProgress();

    return ele.footprint() * refl;
}
