//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      Sim/Simulation/DepthprobeSimulation.cpp
//! @brief     Implements class DepthprobeSimulation
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "Sim/Simulation/DepthprobeSimulation.h"
#include "Base/Axis/Bin.h"
#include "Base/Axis/FixedBinAxis.h"
#include "Base/Math/Constants.h"
#include "Base/Util/Assert.h"
#include "Device/Beam/IFootprintFactor.h"
#include "Device/Coord/CoordSystem2D.h"
#include "Device/Data/Datafield.h"
#include "Device/Histo/SimulationResult.h"
#include "Param/Distrib/DistributionHandler.h"
#include "Param/Distrib/Distributions.h"
#include "Resample/Element/DepthprobeElement.h"
#include "Sim/Computation/DepthprobeComputation.h"
#include "Sim/Scan/AlphaScan.h"
#include <iostream>

DepthprobeSimulation::DepthprobeSimulation(const IBeamScan& scan, const MultiLayer& sample,
                                           const IAxis& zaxis)
    : ISimulation(sample)
    , m_scan(dynamic_cast<AlphaScan*>(scan.clone()))
    , m_z_axis(zaxis.clone())
{
    if (!m_scan)
        throw std::runtime_error("DepthprobeSimulation not implemented for non-alpha scan");
}

DepthprobeSimulation::~DepthprobeSimulation() = default;

const ICoordSystem* DepthprobeSimulation::simCoordSystem() const
{
    std::vector<const IAxis*> axes({m_scan->coordinateAxis()->clone(), m_z_axis->clone()});
    return new DepthprobeCoords(axes, M_TWOPI / m_scan->wavelength()); // TODO mv conversion
}

std::vector<const INode*> DepthprobeSimulation::nodeChildren() const
{
    std::vector<const INode*> result = ISimulation::nodeChildren();
    result.push_back(m_scan.get());
    return result;
}

//... Overridden executors:

//! init callbacks for setting the parameter values
void DepthprobeSimulation::initDistributionHandler()
{
    for (const auto& distribution : distributionHandler().paramDistributions()) {

        switch (distribution.whichParameter()) {
        case ParameterDistribution::BeamInclinationAngle: {
            distributionHandler().defineCallbackForDistribution(
                &distribution, [&](double d) { m_scan->setAlphaOffset(d); });
            break;
        }
        case ParameterDistribution::BeamWavelength:
            distributionHandler().defineCallbackForDistribution(
                &distribution, [&](double d) { m_scan->setWavelength(d); });
            break;
        default:
            ASSERT(false);
        }
    }
}

void DepthprobeSimulation::runComputation(const ReSample& re_sample, size_t i, double weight)
{
    const double result_angle = m_scan->coordinateAxis()->bin(i).center() + m_scan->alphaOffset();
    DepthprobeElement ele(m_scan->wavelength(), -result_angle, m_z_axis.get(),
                          0 < result_angle && result_angle < M_PI_2);

    DepthprobeComputation(re_sample, options(), progress()).run(ele);

    const double alpha_i = -ele.alphaI();
    const auto* footprint = m_scan->footprint();
    double intensity_factor = m_scan->intensity();
    if (footprint != nullptr)
        intensity_factor *= footprint->calculate(alpha_i);
    ele.setIntensities(ele.getIntensities() * intensity_factor);

    if (background())
        throw std::runtime_error(
            "Error: nonzero background is not supported by DepthprobeSimulation");

    const size_t N1 = m_z_axis->size();
    for (size_t j = 0; j < N1; ++j)
        m_cache[i * N1 + j] += ele.getIntensities()[j] * weight;
}

//... Overridden getters:

size_t DepthprobeSimulation::nElements() const
{
    return m_scan->coordinateAxis()->size();
}

size_t DepthprobeSimulation::nOutChannels() const
{
    return nElements() * m_z_axis->size();
}

SimulationResult DepthprobeSimulation::packResult()
{
    const std::vector<const IAxis*> axes{m_scan->coordinateAxis()->clone(), m_z_axis->clone()};
    auto data = std::make_unique<Datafield>(axes, m_cache);

    return {*data, simCoordSystem()};
}
