#include "Base/Axis/FixedBinAxis.h"
#include "Base/Axis/PointwiseAxis.h"
#include "Device/Data/Datafield.h"
#include "Device/IO/IOFactory.h"
#include "GUI/Model/Data/DataItem.h"
#include "GUI/Model/Data/RealItem.h"
#include "GUI/Model/Device/InstrumentItems.h"
#include "GUI/Model/Device/InstrumentModel.h"
#include "GUI/Model/Model/RealModel.h"
#include "GUI/Support/IO/ImportDataInfo.h"
#include "GUI/View/Loaders/AutomaticDataLoader1D.h"
#include "GUI/View/Loaders/DataLoaderUtil.h"
#include "GUI/View/Loaders/QREDataLoader.h"
#include "Tests/GTestWrapper/google_test.h"
#include "Tests/Unit/GUI/Utils.h"
#include <QTest>
#include <fstream>

const QString dir_test_real_model = "TestRealModel";

TEST(TestRealModel, RealItemWithNativeData)
{
    RealModel realModel;

    // initial state
    auto dataItems = realModel.dataItems();
    EXPECT_EQ(dataItems.size(), 0);

    // add 1D RealItem
    auto* realItem1D = realModel.insertSpecularDataItem();
    realItem1D->setRealItemName("RealItem1D");
    EXPECT_EQ(realModel.dataItems().size(), 1);

    // import 1D data with importer
    Datafield data1D({new FixedBinAxis("x", 10, 0.0, 3.0)});
    ImportDataInfo import_data_1D(std::unique_ptr<Datafield>(data1D.clone()), "nbins");
    realItem1D->setImportData(std::move(import_data_1D));

    EXPECT_FALSE(realItem1D->isIntensityData());
    EXPECT_EQ(realModel.dataItems().size(), 2);

    // add 2D RealItem
    auto* realItem2D = realModel.insertIntensityDataItem();
    realItem2D->setRealItemName("RealItem2D");
    EXPECT_EQ(realModel.dataItems().size(), 3);

    // import 2D data with importer
    Datafield data2D({new FixedBinAxis("x", 10, 0.0, 1.0), new FixedBinAxis("y", 10, -1.0, 1.0)});
    ImportDataInfo import_data_2D(std::unique_ptr<Datafield>(data2D.clone()), "nbins");
    realItem2D->setImportData(std::move(import_data_2D));

    EXPECT_TRUE(realItem2D->isIntensityData());
    EXPECT_EQ(realModel.dataItems().size(), 4);

    // check data items of RealModel
    EXPECT_EQ(realModel.dataItems().indexOf(realItem1D->dataItem()), 0);
    EXPECT_EQ(realModel.dataItems().indexOf(realItem1D->nativeDataItem()), 1);
    EXPECT_EQ(realModel.dataItems().indexOf(realItem2D->dataItem()), 2);
    EXPECT_EQ(realModel.dataItems().indexOf(realItem2D->nativeDataItem()), 3);
}

TEST(TestRealModel, removeNativeData)
{
    RealModel model;

    // add RealItem
    auto* item = model.insertSpecularDataItem();
    EXPECT_EQ(item->nativeDataItem(), nullptr);

    // add native data
    item->initNativeData();
    EXPECT_NE(item->nativeDataItem(), nullptr);
    EXPECT_FALSE(item->hasNativeData());
    EXPECT_EQ(item->nativeDataItem()->datafield(), nullptr);

    // call remove while no output data set yet
    item->removeNativeData();
    EXPECT_NE(item->nativeDataItem(), nullptr);
    EXPECT_FALSE(item->hasNativeData());
    EXPECT_EQ(item->nativeDataItem()->datafield(), nullptr);

    // set datafield
    auto* oData = new Datafield{{new PointwiseAxis("qVector", std::vector<double>({1, 2}))}};
    oData->setVector(std::vector<double>({3, 4}));
    item->nativeDataItem()->setDatafield(oData); // takes ownership of odata
    EXPECT_TRUE(item->hasNativeData());
    EXPECT_NE(item->nativeDataItem()->datafield(), nullptr);

    // call remove with existing output data
    item->removeNativeData();
    EXPECT_NE(item->nativeDataItem(), nullptr);
    EXPECT_FALSE(item->hasNativeData());
    EXPECT_EQ(item->nativeDataItem()->datafield(), nullptr);
}

TEST(TestRealModel, saveNonXMLData)
{
    if (!QFile::exists(dir_test_real_model))
        QDir(".").mkdir(dir_test_real_model);

    const QString dir = dir_test_real_model + "/saveNonXMLData";
    UTest::GUI::create_dir(dir);

    const double value1(1.0), value2(2.0), value3(3.0);

    RealModel model;
    RealItem* item1 = UTest::GUI::createRealData("data1", model, value1);
    RealItem* item2 = UTest::GUI::createRealData("data2", model, value2);

    item1->updateDataFileName();
    item2->updateDataFileName();

    // do saving in the main thread
    item1->dataItem()->setSaveInBackground(false);
    item2->dataItem()->setSaveInBackground(false);

    // save first time
    model.writeDataFiles(dir);
    QTest::qSleep(10);

    // check existence of data on disk
    QString fname1 = dir + "/realdata_data1_0.int.gz";
    QString fname2 = dir + "/realdata_data2_0.int.gz";
    EXPECT_TRUE(QFile::exists(fname1));
    EXPECT_TRUE(QFile::exists(fname2));

    // read data from disk, checking it is the same
    std::unique_ptr<Datafield> dataOnDisk1(IOFactory::readDatafield(fname1.toStdString()));
    std::unique_ptr<Datafield> dataOnDisk2(IOFactory::readDatafield(fname2.toStdString()));
    EXPECT_TRUE(UTest::GUI::isTheSame(*dataOnDisk1, *item1->dataItem()->datafield()));
    EXPECT_TRUE(UTest::GUI::isTheSame(*dataOnDisk2, *item2->dataItem()->datafield()));

    // modify data and save the project
    item2->setDatafield(UTest::GUI::createData(value3).release());
    model.writeDataFiles(dir);
    QTest::qSleep(10);

    EXPECT_TRUE(UTest::GUI::isTheSame(*dataOnDisk1, *item1->dataItem()->datafield()));
    EXPECT_TRUE(UTest::GUI::isTheSame(*dataOnDisk2, *item2->dataItem()->datafield()) == false);

    // check that data on disk has changed
    dataOnDisk2.reset(IOFactory::readDatafield(fname2.toStdString()));
    EXPECT_TRUE(UTest::GUI::isTheSame(*dataOnDisk2, *item2->dataItem()->datafield()));

    // rename RealData and check that file on disk changed the name
    item2->setRealItemName("data2new");
    item2->updateDataFileName();
    model.writeDataFiles(dir);
    QTest::qSleep(10);

    QString fname2new = "./" + dir + "/realdata_data2new_0.int.gz";
    EXPECT_TRUE(QFile::exists(fname2new));
    EXPECT_TRUE(UTest::GUI::isTheSame(fname2new, *item2->dataItem()->datafield()));

    // check that file with old name was removed
    EXPECT_FALSE(QFile::exists(fname2));
}

TEST(TestRealModel, saveXMLData)
{
    if (!QFile::exists(dir_test_real_model))
        QDir(".").mkdir(dir_test_real_model);

    const QString dir = dir_test_real_model + "/saveXMLData";
    UTest::GUI::create_dir(dir);

    InstrumentModel instrument_model;
    RealModel model1(&instrument_model);

    // add specular RealItems with non-default parameters
    auto* spec1 = model1.insertSpecularDataItem();
    spec1->setRealItemName("spec1");
    spec1->setNativeFileName("spec1_file_name");
    spec1->setNativeDataUnits("Radians");
    spec1->setPresentationType("Reflectometry (Graph only)");
    spec1->linkToInstrument(instrument_model.addInstrumentItem<SpecularInstrumentItem>());

    // create AutomaticDataLoader1D for specular RealItem
    register1DDataLoaders();
    auto loader1 = new AutomaticDataLoader1D();
    loader1->setRealDataItem(spec1);
    spec1->setDataLoader(loader1);
    loader1->processContents();

    // add second specular RealItem
    auto* spec2 = model1.insertSpecularDataItem();
    spec2->setRealItemName("spec2");

    // create QREDataLoader for specular RealItem
    auto loader2 = new QREDataLoader();
    loader2->setRealDataItem(spec2);
    spec2->setDataLoader(loader2);
    loader2->processContents();

    // add 2D RealItems with non-default parameters
    auto* intensity1 = model1.insertIntensityDataItem();
    auto* intensity2 = model1.insertIntensityDataItem();
    intensity1->setRealItemName("GISAS");
    intensity2->setRealItemName("OffSpec");
    intensity1->linkToInstrument(instrument_model.addInstrumentItem<GISASInstrumentItem>());
    intensity2->linkToInstrument(instrument_model.addInstrumentItem<OffspecInstrumentItem>());

    // set non-default top-level model parameters
    model1.setSelectedRank(2);
    model1.setSelectedIndex(1);

    // save data to project files
    const QString file1 = dir + "/Project_1.ba";
    const QString file2 = dir + "/Project_2.ba";

    // write changed model to disk
    const QString tag = "RealModel";
    UTest::GUI::writeXMLFile<RealModel>(file1, model1, tag);
    EXPECT_TRUE(QFile::exists(file1));

    // read data to the second model
    RealModel model2(&instrument_model);
    UTest::GUI::readXMLFile<RealModel>(file1, model2, tag);

    // write the second model back to disk
    UTest::GUI::writeXMLFile<RealModel>(file2, model2, tag);
    EXPECT_TRUE(QFile::exists(file2));

    // compare files byte-by-byte
    std::ifstream f1(file1.toStdString());
    std::ifstream f2(file2.toStdString());

    typedef std::istreambuf_iterator<char> it;

    std::vector<char> contents1((it(f1)), it());
    std::vector<char> contents2((it(f2)), it());

    EXPECT_EQ(contents1, contents2);
}
