#include "Device/Coord/CoordSystem1D.h"
#include "Base/Axis/FixedBinAxis.h"
#include "Base/Axis/PointwiseAxis.h"
#include "Base/Axis/VariableBinAxis.h"
#include "Base/Const/Units.h"
#include "Base/Math/Constants.h"
#include "Device/Data/Datafield.h"
#include "Sim/Scan/QzScan.h"
#include "Tests/GTestWrapper/google_test.h"
#include <algorithm>

class CoordSystem1DTest : public ::testing::Test {
public:
    CoordSystem1DTest();

    double getQ(double angle) { return 4.0 * M_PI * std::sin(angle) / m_wavelength; }

protected:
    void checkConventionalConverter(const CoordSystem1D& test_object);
    void checkQSpecConverter(const CoordSystem1D& test_object);
    FixedBinAxis m_axis;
    FixedBinAxis m_q_axis;
    QzScan m_qscan;
    double m_wavelength{1.};
};

CoordSystem1DTest::CoordSystem1DTest()
    : m_axis("Angles", 5, 0.5, 1.0)     // angles in radians
    , m_q_axis("Q values", 5, 0.0, 1.0) // q axis in inv. nm
    , m_qscan(m_q_axis)
{
}

void CoordSystem1DTest::checkConventionalConverter(const CoordSystem1D& test_object)
{
    double expected_min = m_axis.binCenter(0);
    EXPECT_NEAR(test_object.calculateMin(0, Coords::UNDEFINED), Units::rad2deg(expected_min),
                Units::rad2deg(expected_min) * 1e-10);
    EXPECT_NEAR(test_object.calculateMin(0, Coords::NBINS), 0.0, 1e-10);
    EXPECT_NEAR(test_object.calculateMin(0, Coords::RADIANS), expected_min, expected_min * 1e-10);
    EXPECT_NEAR(test_object.calculateMin(0, Coords::DEGREES), Units::rad2deg(expected_min),
                Units::rad2deg(expected_min) * 1e-10);
    EXPECT_NEAR(test_object.calculateMin(0, Coords::QSPACE), getQ(expected_min),
                getQ(expected_min) * 1e-10);
    EXPECT_NEAR(test_object.calculateMin(0, Coords::RQ4), getQ(expected_min),
                getQ(expected_min) * 1e-10);

    double expected_max = m_axis.binCenters().back();
    EXPECT_NEAR(test_object.calculateMax(0, Coords::UNDEFINED), Units::rad2deg(expected_max),
                Units::rad2deg(expected_max) * 1e-10);
    EXPECT_NEAR(test_object.calculateMax(0, Coords::NBINS), static_cast<double>(m_axis.size()),
                1e-10);
    EXPECT_NEAR(test_object.calculateMax(0, Coords::RADIANS), expected_max, expected_max * 1e-10);
    EXPECT_NEAR(test_object.calculateMax(0, Coords::DEGREES), Units::rad2deg(expected_max),
                Units::rad2deg(expected_max) * 1e-10);
    EXPECT_NEAR(test_object.calculateMax(0, Coords::QSPACE), getQ(expected_max),
                getQ(expected_max) * 1e-10);
    EXPECT_NEAR(test_object.calculateMax(0, Coords::RQ4), getQ(expected_max),
                getQ(expected_max) * 1e-10);

    // DEFAULT
    std::unique_ptr<IAxis> axis_default(test_object.createConvertedAxis(0, Coords::UNDEFINED));
    const IAxis* tmp = axis_default.get();
    EXPECT_TRUE(dynamic_cast<const PointwiseAxis*>(tmp));
    EXPECT_EQ(axis_default->size(), test_object.axisSize(0));
    EXPECT_EQ(axis_default->min(), test_object.calculateMin(0, Coords::UNDEFINED));
    EXPECT_EQ(axis_default->max(), test_object.calculateMax(0, Coords::UNDEFINED));

    // QSPACE
    std::unique_ptr<const IAxis> axis_qspace(test_object.createConvertedAxis(0, Coords::QSPACE));
    EXPECT_TRUE(dynamic_cast<const PointwiseAxis*>(axis_qspace.get()));
    EXPECT_EQ(axis_qspace->size(), test_object.axisSize(0));
    EXPECT_EQ(axis_qspace->min(), test_object.calculateMin(0, Coords::QSPACE));
    EXPECT_EQ(axis_qspace->max(), test_object.calculateMax(0, Coords::QSPACE));

    // NBINS
    std::unique_ptr<IAxis> axis_nbins(test_object.createConvertedAxis(0, Coords::NBINS));
    EXPECT_TRUE(dynamic_cast<const FixedBinAxis*>(axis_nbins.get()));
    EXPECT_EQ(axis_nbins->size(), test_object.axisSize(0));
    EXPECT_EQ(axis_nbins->min(), test_object.calculateMin(0, Coords::NBINS));
    EXPECT_EQ(axis_nbins->max(), test_object.calculateMax(0, Coords::NBINS));

    // RQ4
    std::unique_ptr<IAxis> axis_rq4(test_object.createConvertedAxis(0, Coords::RQ4));
    EXPECT_TRUE(dynamic_cast<const PointwiseAxis*>(axis_rq4.get()));
    EXPECT_EQ(axis_rq4->size(), test_object.axisSize(0));
    EXPECT_EQ(axis_rq4->min(), test_object.calculateMin(0, Coords::RQ4));
    EXPECT_EQ(axis_rq4->max(), test_object.calculateMax(0, Coords::RQ4));
    EXPECT_TRUE(*axis_rq4 == *axis_qspace);

    Datafield fake_data({m_axis.clone()});
    std::vector<double> raw_fake(m_axis.size(), 1.0);
    fake_data.setVector(raw_fake);

    // NBINS
    auto data_nbins = test_object.createConvertedData(fake_data, Coords::NBINS);
    EXPECT_EQ(data_nbins->axis(0), *axis_nbins);
    EXPECT_EQ(data_nbins->size(), axis_nbins->size());
    EXPECT_EQ(raw_fake, data_nbins->flatVector());

    // RQ4
    auto data_rq4 = test_object.createConvertedData(fake_data, Coords::RQ4);
    EXPECT_EQ(data_rq4->axis(0), *axis_rq4);
    auto raw_data_rq4 = data_rq4->flatVector();
    EXPECT_EQ(raw_data_rq4.size(), raw_fake.size());
    EXPECT_EQ(raw_data_rq4.size(), axis_rq4->size());
    for (size_t i = 0, size = raw_data_rq4.size(); i < size; ++i) {
        double value = raw_fake[i] * std::pow(axis_rq4->binCenter(i), 4);
        EXPECT_DOUBLE_EQ(raw_data_rq4[i], value);
    }
}

void CoordSystem1DTest::checkQSpecConverter(const CoordSystem1D& test_object)
{
    double expected_min = m_q_axis.binCenter(0);
    EXPECT_EQ(test_object.calculateMin(0, Coords::UNDEFINED), expected_min);
    EXPECT_NEAR(test_object.calculateMin(0, Coords::NBINS), 0.0, 1e-10);
    EXPECT_EQ(test_object.calculateMin(0, Coords::QSPACE), expected_min);
    EXPECT_EQ(test_object.calculateMin(0, Coords::RQ4), expected_min);

    double expected_max = m_q_axis.binCenters().back();
    EXPECT_EQ(test_object.calculateMax(0, Coords::UNDEFINED), expected_max);
    EXPECT_NEAR(test_object.calculateMax(0, Coords::NBINS), static_cast<double>(m_q_axis.size()),
                1e-10);
    EXPECT_EQ(test_object.calculateMax(0, Coords::QSPACE), expected_max);
    EXPECT_EQ(test_object.calculateMax(0, Coords::RQ4), expected_max);

    // DEFAULT
    std::unique_ptr<IAxis> axis_default(test_object.createConvertedAxis(0, Coords::UNDEFINED));
    EXPECT_TRUE(dynamic_cast<const PointwiseAxis*>(axis_default.get()));
    EXPECT_EQ(axis_default->size(), test_object.axisSize(0));
    EXPECT_EQ(axis_default->min(), test_object.calculateMin(0, Coords::UNDEFINED));
    EXPECT_EQ(axis_default->max(), test_object.calculateMax(0, Coords::UNDEFINED));

    // QSPACE
    std::unique_ptr<IAxis> axis_qspace(test_object.createConvertedAxis(0, Coords::QSPACE));
    EXPECT_TRUE(dynamic_cast<const PointwiseAxis*>(axis_qspace.get()));
    EXPECT_EQ(axis_qspace->size(), test_object.axisSize(0));
    EXPECT_EQ(axis_qspace->min(), test_object.calculateMin(0, Coords::QSPACE));
    EXPECT_EQ(axis_qspace->max(), test_object.calculateMax(0, Coords::QSPACE));
    EXPECT_EQ(*axis_default, *axis_qspace);

    // NBINS
    std::unique_ptr<IAxis> axis_nbins(test_object.createConvertedAxis(0, Coords::NBINS));
    EXPECT_TRUE(dynamic_cast<const FixedBinAxis*>(axis_nbins.get()));
    EXPECT_EQ(axis_nbins->size(), test_object.axisSize(0));
    EXPECT_EQ(axis_nbins->min(), test_object.calculateMin(0, Coords::NBINS));
    EXPECT_EQ(axis_nbins->max(), test_object.calculateMax(0, Coords::NBINS));

    // RQ4
    std::unique_ptr<IAxis> axis_rq4(test_object.createConvertedAxis(0, Coords::RQ4));
    EXPECT_TRUE(dynamic_cast<const PointwiseAxis*>(axis_rq4.get()));
    EXPECT_EQ(axis_rq4->size(), test_object.axisSize(0));
    EXPECT_EQ(axis_rq4->min(), test_object.calculateMin(0, Coords::RQ4));
    EXPECT_EQ(axis_rq4->max(), test_object.calculateMax(0, Coords::RQ4));
    EXPECT_TRUE(*axis_rq4 == *axis_qspace);

    Datafield fake_data({m_axis.clone()});
    std::vector<double> raw_fake(m_axis.size(), 1.0);
    fake_data.setVector(raw_fake);

    // NBINS
    auto data_nbins = test_object.createConvertedData(fake_data, Coords::NBINS);
    EXPECT_EQ(data_nbins->axis(0), *axis_nbins);
    EXPECT_EQ(data_nbins->size(), axis_nbins->size());
    EXPECT_EQ(raw_fake, data_nbins->flatVector());

    // RQ4
    auto data_rq4 = test_object.createConvertedData(fake_data, Coords::RQ4);
    EXPECT_EQ(data_rq4->axis(0), *axis_rq4);
    auto raw_data_rq4 = data_rq4->flatVector();
    EXPECT_EQ(raw_data_rq4.size(), raw_fake.size());
    EXPECT_EQ(raw_data_rq4.size(), axis_rq4->size());
    for (size_t i = 0, size = raw_data_rq4.size(); i < size; ++i) {
        double value = raw_fake[i] * std::pow(axis_rq4->binCenter(i), 4);
        EXPECT_DOUBLE_EQ(raw_data_rq4[i], value);
    }
}

TEST_F(CoordSystem1DTest, MainFunctionality)
{
    checkConventionalConverter(AngularReflectometryCoords(m_wavelength, m_axis));
    checkQSpecConverter(WavenumberReflectometryCoords(m_qscan.coordinateAxis()->clone()));
}

TEST_F(CoordSystem1DTest, Exceptions)
{
    AngularReflectometryCoords converter(m_wavelength, m_axis);

    EXPECT_FAILED_ASSERT(converter.calculateMin(0, Coords::MM));
    EXPECT_FAILED_ASSERT(converter.calculateMin(1, Coords::RADIANS));

    EXPECT_FAILED_ASSERT(converter.calculateMax(0, Coords::MM));
    EXPECT_FAILED_ASSERT(converter.calculateMax(1, Coords::RADIANS));

    EXPECT_FAILED_ASSERT(converter.createConvertedAxis(0, Coords::MM));
    EXPECT_FAILED_ASSERT(converter.createConvertedAxis(1, Coords::UNDEFINED));

    FixedBinAxis axis("Angles", 100, 0.0, M_TWOPI);
    EXPECT_FAILED_ASSERT(AngularReflectometryCoords converter2(m_wavelength, axis));

    WavenumberReflectometryCoords converter2(m_qscan.coordinateAxis()->clone());
    // wrong units
    EXPECT_FAILED_ASSERT(converter2.calculateMin(0, Coords::MM));
    EXPECT_FAILED_ASSERT(converter2.calculateMin(0, Coords::RADIANS));
    EXPECT_FAILED_ASSERT(converter2.calculateMin(0, Coords::DEGREES));
    EXPECT_FAILED_ASSERT(converter2.calculateMax(0, Coords::MM));
    EXPECT_FAILED_ASSERT(converter2.calculateMax(0, Coords::RADIANS));
    EXPECT_FAILED_ASSERT(converter2.calculateMax(0, Coords::DEGREES));
    EXPECT_FAILED_ASSERT(converter2.createConvertedAxis(0, Coords::MM));

    // wrong axis index
    EXPECT_FAILED_ASSERT(converter2.calculateMin(1, Coords::QSPACE));
    EXPECT_FAILED_ASSERT(converter2.calculateMax(1, Coords::RQ4));
    EXPECT_FAILED_ASSERT(converter2.createConvertedAxis(1, Coords::UNDEFINED));
}

TEST_F(CoordSystem1DTest, Clone)
{
    AngularReflectometryCoords converter(m_wavelength, m_axis);
    std::unique_ptr<CoordSystem1D> converter_clone(converter.clone());
    checkConventionalConverter(*converter_clone);

    WavenumberReflectometryCoords converterQ(m_qscan.coordinateAxis()->clone());
    std::unique_ptr<CoordSystem1D> converterQ_clone(converterQ.clone());
    checkQSpecConverter(*converterQ_clone);
}

TEST_F(CoordSystem1DTest, NonDefaultUnitsInInput)
{
    PointwiseAxis axis("x", std::vector<double>{0.0, 0.5, 1.0});

    EXPECT_FAILED_ASSERT(AngularReflectometryCoords(m_wavelength, axis, Coords::NBINS));

    AngularReflectometryCoords converter(m_wavelength, axis, Coords::DEGREES);
    auto axis_deg_output = converter.createConvertedAxis(0, Coords::DEGREES);
    EXPECT_TRUE(axis.size() == axis_deg_output->size());
    EXPECT_DOUBLE_EQ(axis[0], (*axis_deg_output)[0]);
    EXPECT_DOUBLE_EQ(axis[1], (*axis_deg_output)[1]);
    EXPECT_DOUBLE_EQ(axis[2], (*axis_deg_output)[2]);

    auto values = axis.binCenters();
    std::for_each(values.begin(), values.end(), [this](double& value) { value = getQ(value); });
    PointwiseAxis q_axis("q", values);
    AngularReflectometryCoords converter2(m_wavelength, q_axis, Coords::QSPACE);
    auto axis_rad_output = converter2.createConvertedAxis(0, Coords::RADIANS);
    EXPECT_TRUE(axis.size() == axis_rad_output->size());
    EXPECT_DOUBLE_EQ(axis[0], (*axis_rad_output)[0]);
    EXPECT_DOUBLE_EQ(axis[1], (*axis_rad_output)[1]);
    EXPECT_DOUBLE_EQ(axis[2], (*axis_rad_output)[2]);

    EXPECT_FAILED_ASSERT(AngularReflectometryCoords(m_wavelength, q_axis, Coords::RQ4));
}
